from datetime import datetime
from typing import List, Optional

import pytest
from maypy.predicates import contains, is_length

from power_events.conditions import Neg
from power_events.conditions.value import ABSENT, Value, ValuePath, combine
from power_events.exceptions import NoPredicateError, ValueAbsentError


def test_combine() -> None:
    pred = combine(is_length(2), contains("maypy"))

    assert pred(["power-events", "maypy"])
    assert not pred(["mapy"])
    assert not pred([1, 2])


class TestValuePath:
    def test_should_raise_error_when_path_begin_by_sep(self) -> None:
        with pytest.raises(ValueError, match="Path value should not begin"):
            ValuePath(".a.b")

    def test_should_raise_error_when_path_end_by_sep(self) -> None:
        with pytest.raises(ValueError, match="Path value should not end"):
            ValuePath("a.b.")

    def test_should_raise_error_when_path_has_consecutive_sep(self) -> None:
        with pytest.raises(ValueError, match="Path value should not contain consecutive"):
            ValuePath("a..b")

    @pytest.mark.parametrize(
        ("path", "sep", "keys"),
        [
            ("", None, []),
            ("effective-date/1", "/", ["effective-date", "1"]),
            ("effective+date+1", "+", ["effective", "date", "1"]),
        ],
    )
    def test_should_create_when_valid_path(
        self, path: str, sep: Optional[str], keys: List[str]
    ) -> None:
        assert ValuePath(path, separator=sep).keys == keys

    def test_get_when_path_empty_should_return_event(self) -> None:
        assert ValuePath("").get({"a": 1}) == {"a": 1}

    def test_get_should_return_value(self) -> None:
        assert ValuePath("a.b").get_from({"a": {"b": 2}}) == 2

    def test_get_should_handle_int_key(self) -> None:
        assert ValuePath("a.b.1").get_from({"a": {"b": {1: "int"}}}) == "int"
        assert ValuePath("a.b.1").get_from({"a": {"b": {"1": "str"}}}) == "str"

    def test_get_should_return_absent_sentinel_by_default_when_no_value_at_path(self) -> None:
        assert ValuePath("a.b").get_from({"a": {"c": 2}}) is ABSENT

    def test_get_should_return_default_value_when_given_and_value_absent(self) -> None:
        path = ValuePath("a.b")
        assert path.get_from({}, None) is None
        assert path.get_from({"a": 1}, None) is None

    def test_get_should_raise_error_when_no_value_and_flag_set(self) -> None:
        with pytest.raises(ValueAbsentError) as excinfo:
            ValuePath("foo.bar").get_from({"foo": {}}, raise_if_absent=True)
            assert excinfo.value.missing_key == "bar"
            assert excinfo.value.path == "foo.bar"


class TestValue:
    def test_root(self) -> None:
        assert Value.root().contains("a").check({"a": 1})

    def test_is_truthy(self) -> None:
        assert Value("a.b").is_truthy().check({"a": {"b": True}})
        assert not Value("a.b").is_truthy().check({"a": {"b": False}})

    def test_equal_with_none(self) -> None:
        assert Value("a").equals(None).check({"a": None})

    def test_contains(self) -> None:
        assert Value("a.b").contains(1, 2).check({"a": {"b": [1, 2]}})

    def test_is_not_empty(self) -> None:
        assert Value("a.b").is_not_empty().check({"a": {"b": "not empty"}})
        assert not Value("a.b").is_not_empty().check({"a": {"b": []}})

    def test_is_size(self) -> None:
        assert Value("a.b").is_length(3).check({"a": {"b": [1, 2, 3]}})
        assert not Value("a.b").is_length(3).check({"a": {"b": []}})

    def test_one_of(self) -> None:
        assert Value("a.b").one_of(["foo", "bar"]).check({"a": {"b": "bar"}})
        assert not Value("a.b").one_of(["foo", "bar"]).check({"a": {"b": "baz"}})

    def test_match_regex(self) -> None:
        assert Value("a.b").match_regex(r"test_\d{2,}").check({"a": {"b": "test_123"}})
        assert not Value("a.b").match_regex(r"test_\d{2,}").check({"a": {"b": "test"}})

    def test_match(self) -> None:
        def is_even(val: int) -> bool:
            return val % 2 == 0

        assert Value("a.b").match(is_even).check({"a": {"b": 8}})
        assert not Value("a.b").match(is_even).check({"a": {"b": 7}})

    def test_equals(self) -> None:
        assert not Value("a.b.c").equals(2).check({"a": {"b": {"c": 1}}})

    def test_check_value(self) -> None:
        assert Value("a.b").equals(2).check({"a": {"b": 2}})

    def test_invert(self) -> None:
        value = Value("a.b").equals(2)

        assert ~value.check({"a": {"b": 1}})
        assert Neg(value).check({"a": {"b": 1}})

    def test_check_should_raise_error_when_no_predicate_set(self) -> None:
        with pytest.raises(NoPredicateError):
            Value("a.b.c").check({})

    def test_check_raise_error_when_value_not_in_event_and_flag_raise(self) -> None:
        with pytest.raises(ValueAbsentError):
            Value("a.b.c").equals(2).check({}, raise_if_absent=True)

    def test_check_should_return_false_when_value_absent(self) -> None:
        assert not Value("a.b.c").equals(2).check({})

    def test_check_multiple_predicates(self) -> None:
        value = Value("a").contains(1).is_length(3)
        assert value.check({"a": [3, 2, 1]})
        assert not value.check({"a": [3, 1]})
        assert not value.check({"a": [3, 2, 4]})

    def test_mapper(self) -> None:
        value = Value("a").match(lambda val: val < datetime.now())

        with pytest.raises(TypeError):
            value.check({"a": "2021-02-08"})

        value.mapper = lambda val: datetime.strptime(val, "%Y-%m-%d")

        assert value.check({"a": "2021-02-08"})



from collections.abc import Sequence
from dataclasses import dataclass
from logging import Logger
from typing import (
    Any,
    Callable,
    Container,
    Dict,
    List,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from maypy.predicates import is_empty
from typing_extensions import Concatenate, ParamSpec

from .conditions import Condition, Value
from .exceptions import MultipleRoutesError, NoRouteFoundError

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")
Error = TypeVar("Error", bound=Exception)
P = ParamSpec("P")

Func = Callable[Concatenate[Dict[Any, Any], P], Any]

logger = Logger("power_events")


@dataclass(frozen=True)
class EventRoute:
    """Class representing an event route with a condition and a function."""

    func: Callable[..., Any]
    condition: Condition

    def match(self, event: Mapping[str, V]) -> bool:
        """Check if the event matches the route's condition.

        Args:
            event: The event to check.
        """
        return self.condition.check(event)

    @property
    def name(self) -> str:
        """Get the name of the route function."""
        return self.func.__name__


class EventResolver:
    """Class responsible for resolving events against registered routes."""

    def __init__(self, allow_multiple_routes: bool = False, allow_no_route: bool = True) -> None:
        """Initialize the event resolver with optional configuration.

        Args:
            allow_multiple_routes: option to allow multiples routes on same event, otherwise raise `MultipleRoutesError`.
            allow_no_route: option to allow no routes on event, otherwise raise `NoRouteFoundError`.
        """
        self._routes: List[EventRoute] = []
        self._fallback_route: Optional[EventRoute] = None
        self._exception_handlers: Dict[Type[Exception], Callable[..., Any]] = {}
        self._allow_multiple_routes = allow_multiple_routes
        self._allow_no_route = allow_no_route

    def equal(self, value_path: str, expected: Any) -> Callable[[Func[P]], Func[P]]:
        """Register a route with an equality condition.

        Args:
            value_path: The path to the value in the event.
            expected: The expected value.
        """
        return self.when(Value(value_path).equals(expected))

    def one_of(self, value_path: str, options: Container[V]) -> Callable[[Func[P]], Func[P]]:
        """Register a route with a one-of condition.

        Args:
            value_path: The path to the value in the event.
            options: The container of expected values.
        """
        return self.when(Value(value_path).one_of(options))

    def contain(self, value_path: str, *items: V) -> Callable[[Func[P]], Func[P]]:
        """Register a route where value should contain items.

        Args:
            value_path: The path to the value in the event.
            items: Items to in the event.
        """
        return self.when(Value(value_path).contains(*items))

    def when(self, condition: Condition) -> Callable[[Func[P]], Func[P]]:
        """Register a route with a custom condition.

        Args:
            condition: The condition to trigger this route.
        """

        def register_route(fn: Func[P]) -> Func[P]:
            route = EventRoute(condition=condition, func=fn)
            self._routes.append(route)
            return fn

        return register_route

    @overload
    def exception_handler(
        self, exc_type: Type[Error]
    ) -> Callable[[Callable[[Error], Any]], Callable[[Error], Any]]: ...

    @overload
    def exception_handler(
        self, exc_type: Sequence[Type[Exception]]
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...

    def exception_handler(
        self, exc_type: Union[Type[Error], Sequence[Type[Error]]]
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """Register a function to handle certain exception type.

        Parameters:
            exc_type: exception type to handle.
                Could be either a unique exception type or a list of types.
        """

        def register_exception(fn: Callable[..., Any]) -> Callable[..., Any]:
            if isinstance(exc_type, Sequence):
                for exc in exc_type:
                    self._exception_handlers[exc] = fn
            else:
                self._exception_handlers[exc_type] = fn
            return fn

        return register_exception

    @overload
    def fallback(self) -> Callable[[Func[P]], Func[P]]: ...

    @overload
    def fallback(self, func: Func[P]) -> Func[P]: ...

    def fallback(
        self, func: Optional[Func[P]] = None
    ) -> Union[Callable[[Func[P]], Func[P]], Func[P]]:
        """Register a fallback route if no registered routes match the event."""

        def register_fallback(fn: Func[P]) -> Func[P]:
            self._fallback_route = EventRoute(fn, Value.root().match(lambda x: True))
            return fn

        if func is None:
            return register_fallback  # pragma: no cover

        return register_fallback(func)

    def resolve(self, event: Mapping[Any, V]) -> Sequence[Any]:
        """Resolve the event to the matching routes and execute their functions.

        Args:
            event: The event to resolve.
        """
        available_routes = self._find_matching_routes(event)
        try:
            self._handle_not_found(event, available_routes)
            self._handle_multiple_routes(event, available_routes)

            return [route.func(event) for route in available_routes]

        except Exception as exc:
            handler = self._lookup_exception_handler(exc)
            if handler:
                return [handler(exc)]

            raise

    def _find_matching_routes(self, event: Mapping[Any, V]) -> List[EventRoute]:
        """Find the routes matching the event."""
        matching_routes = [route for route in self._routes if route.match(event)]

        if is_empty(matching_routes) and self._fallback_route:
            logger.debug("Use fallback route.")
            return [self._fallback_route]

        return matching_routes

    def _handle_not_found(self, event: Mapping[Any, V], available_routes: List[EventRoute]) -> None:
        """Handle cases where no routes match the event.

        Args:
            event: The event to resolve.
            available_routes: The list of matching routes.

        Raises:
            NoRouteFoundError: If no routes are found and not allowed.
        """
        if is_empty(available_routes):
            if not self._allow_no_route:
                raise NoRouteFoundError(event, [route.name for route in self._routes])

            logger.warning("No routes for this event")  # pragma: no cover

    def _handle_multiple_routes(
        self, event: Mapping[Any, V], available_routes: List[EventRoute]
    ) -> None:
        """Handle cases where multiple routes match the event.

        Args:
            event: The event to resolve.
            available_routes: The list of matching routes.

        Raises:
            MultipleRoutesError: If multiple routes are found and not allowed.
        """
        if len(available_routes) > 1:
            if not self._allow_multiple_routes:
                raise MultipleRoutesError(event, [route.name for route in available_routes])
            logger.warning("Multiple routes for this event")  # pragma: no cover

    def _lookup_exception_handler(self, exc: Exception) -> Optional[Callable[[Exception], Any]]:
        """Lookup the handler for the exception using Method Resolution Order, for matching against base exception."""
        for cls in type(exc).__mro__:
            if cls in self._exception_handlers:
                return self._exception_handlers[cls]

        return None



"""Top-level Power event package."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = "undefined"

from .resolver import EventResolver

__all__ = ["EventResolver"]



"""Top-level conditions package.

This package contains all the logic about the condition.
"""

from power_events.conditions.condition import And, Condition, Neg, Or
from power_events.conditions.value import Value, ValuePath

__all__ = [
    "And",
    "Condition",
    "Neg",
    "Or",
    "Value",
    "ValuePath",
]



#!/usr/bin/env python
import re
import sys
from pathlib import Path
from string import Template

CHANGELOG_PATH = Path("CHANGELOG.md")
LATEST_CHANGES_PATTERN = re.compile(
    r"(## (?P<version>\d+\.\d+\.\d+) \(\d{4}-\d{2}-\d{2}\)\n\n(?P<content>.+?))(?=\n\n## |\Z)",
    re.DOTALL,
)

CHANGE_TEMPLATE = Template(
    """
# $version

## What's changed

$content
    """
)


def extract_latest_changes() -> str:
    """Extract latest version changes to create release message."""
    changelog_content = CHANGELOG_PATH.read_text()
    latest_changes_match = LATEST_CHANGES_PATTERN.search(changelog_content)

    if not latest_changes_match:
        sys.exit(-1)

    version = latest_changes_match.group("version")
    content = latest_changes_match.group("content")

    return CHANGE_TEMPLATE.substitute(version=version, content=content)


if __name__ == "__main__":
    print(extract_latest_changes())



from power_events.conditions.condition import (
    And,
    Or,
)
from power_events.conditions.value import Value


class TestCondition:
    def test_condition_or(self) -> None:
        condition = Value("a.b").equals(1) | Value("a.c").equals("maypy")

        assert condition.check({"a": {"b": 0, "c": "maypy"}})

    def test_condition_or_no_match(self) -> None:
        condition = Value("a.b").equals(1) | Value("a.c").equals("maypy")

        assert not condition.check({"a": {"b": 0, "c": 1}})

    def test_condition_and(self) -> None:
        condition = Value("a.b").equals(1) & Value("a.c").equals("maypy")

        assert condition.check({"a": {"b": 1, "c": "maypy"}})

    def test_condition_and_no_match(self) -> None:
        condition = Value("a.b").equals(1) & Value("a.c").equals("maypy")

        assert not condition.check({"a": {"b": 1, "c": 2}})

    def test_nested_or_conditions(self) -> None:
        condition = (Value("a.b").equals(1) & Value("a.c").equals("maypy")) | Value("d").is_truthy()

        assert condition.check({"a": {"b": 1, "c": "maypy"}, "d": False})
        assert condition.check({"a": {"b": 1, "c": "maypy"}, "d": True})
        assert condition.check({"a": {"b": 2, "c": "maypy"}, "d": True})
        assert not condition.check({"a": {"b": 2, "c": "maypy"}, "d": False})

    def test_nested_and_conditions(self) -> None:
        condition = (Value("a.b").equals(1) & Value("a.c").equals("maypy")) & Value("d").is_truthy()

        assert condition.check({"a": {"b": 1, "c": "maypy"}, "d": True})
        assert not condition.check({"a": {"b": 1, "c": "maypy"}, "d": False})
        assert not condition.check({"a": {"b": 2, "c": "maypy"}, "d": True})
        assert not condition.check({"a": {"b": 2, "c": "maypy"}, "d": False})

    def test_invert_or_condition(self) -> None:
        condition = ~(Value("a.b").equals(1) | Value("a.c").equals("maypy"))

        assert isinstance(condition, And)

        assert condition.check({"a": {"b": 0, "c": 1}})
        assert not condition.check({"a": {"b": 1, "c": 1}})
        assert not condition.check({"a": {"b": 0, "c": "maypy"}})
        assert not condition.check({"a": {"b": 1, "c": "maypy"}})

    def test_invert_and_condition(self) -> None:
        condition = ~(Value("a.b").equals(1) & Value("a.c").equals("maypy"))

        assert isinstance(condition, Or)

        assert condition.check({"a": {"b": 0, "c": "maypy"}})
        assert condition.check({"a": {"b": 1, "c": "power-events"}})
        assert not condition.check({"a": {"b": 1, "c": "maypy"}})



from abc import ABC, abstractmethod
from enum import Enum
from typing import (
    Any,
    Callable,
    Iterable,
    Mapping,
    TypeVar,
)

from typing_extensions import override

K = TypeVar("K")
V = TypeVar("V")
Event = Mapping[Any, V]


class Condition(ABC):
    """Abstract base class for a condition that can be checked against an event."""

    @abstractmethod
    def check(self, event: Event[V]) -> bool:
        """Check if the condition holds for the given event.

        Args:
            event: The event to check.
        """

    @abstractmethod
    def __or__(self, other: "Condition") -> "Condition":
        """Combine this condition with another condition using OR logic.

        Args:
            other: The other condition.
        """

    @abstractmethod
    def __and__(self, other: "Condition") -> "Condition":
        """Combine this condition with another condition using AND logic.

        Args:
            other: The other condition.
        """

    @abstractmethod
    def __invert__(self) -> "Condition":
        """Invert this condition (logical NOT)."""


class ConditionOperator(Enum):
    """Enum representing logical operators for condition combination."""

    AND = all
    OR = any

    def __call__(self, *condition: Iterable[Any]) -> bool:
        """Apply the logical operator to the given conditions.

        Args:
            condition: The conditions to combine.
        """
        operator: Callable[[Iterable[Any]], bool] = self.value
        return operator(*condition)


class ConditionExpression(Condition, ABC):
    """Abstract base class for a composite condition expression."""

    operator: ConditionOperator

    def __init__(self, *conditions: Condition) -> None:
        """Initialize the composite condition with the specified conditions.

        Args:
            conditions: The conditions that make up the expression.
        """
        self._conditions = conditions

    @override
    def check(self, event: Event[V]) -> bool:
        return self.operator(condition.check(event) for condition in self._conditions)

    @override
    def __and__(self, other: Condition) -> Condition:
        if not isinstance(other, Condition):
            return NotImplemented
        return And(self, other)

    @override
    def __or__(self, other: Condition) -> Condition:
        if not isinstance(other, Condition):
            return NotImplemented
        return Or(self, other)


class Or(ConditionExpression):
    """Composite condition representing the logical OR of its conditions."""

    operator = ConditionOperator.OR

    @override
    def __invert__(self) -> Condition:
        return And(*(~condition for condition in self._conditions))


class And(ConditionExpression):
    """Composite condition representing the logical AND of its conditions."""

    operator = ConditionOperator.AND

    @override
    def __invert__(self) -> Condition:
        return Or(*(~condition for condition in self._conditions))


def Neg(condition: Condition) -> Condition:
    """Create a new condition representing the logical NOT of the given condition.

    Args:
        condition: The condition to invert.
    """
    return ~condition



from typing import Any, List, Mapping


class PowerEventsError(Exception):
    """Base class for exceptions of power_events."""


class ValueAbsentError(PowerEventsError):
    """Exception raised when the value to check is not present inside the event."""

    def __init__(self, path: str, missing_key: str, event: Mapping[Any, Any]) -> None:
        self.path = path
        self.missing_key = missing_key
        super().__init__(
            f"The value define by path <{path}> is missing the key <{missing_key}> in event:\n{event}"
        )


class NoPredicateError(PowerEventsError):
    """Exception raised when no predicate has been set for a value condition."""

    def __init__(self, path: str) -> None:
        """Initialize the exception with the path that caused the error.

        Args:
            path: The path of the value condition.
        """
        super().__init__(f"No predicate has been set for the value condition on path {path}")


class RouteError(PowerEventsError):
    """Base class for exceptions related to routing errors."""


class NoRouteFoundError(RouteError):
    """Exception raised when no route is found for an event."""

    def __init__(self, event: Mapping[Any, Any], routes: List[str]) -> None:
        """Initialize the exception with the event and the registered routes.

        Args:
            event: The event that caused the error.
            routes: The list of registered route functions.
        """
        self.registered_routes = routes
        super().__init__(
            f"No route found for the current event: {event}.\n"
            f"Registered route functions are {', '.join(routes)}.\n"
            "If it's normal pass the option 'allow_no_route' at `False` in the resolver definition."
        )


class MultipleRoutesError(RouteError):
    """Exception raised when multiple routes are found for an event."""

    def __init__(self, event: Mapping[Any, Any], routes: List[str]) -> None:
        """Initialize the exception with the event and the available routes.

        Args:
            event: The event that caused the error.
            routes: The list of available route functions.
        """
        self.available_routes = routes
        super().__init__(
            f"Multiples routes found for the current event: {event}.\n"
            f"Available route functions are {', '.join(routes)}.\n"
            "If it's normal pass the option 'allow_multiple_routes' in the resolver definition."
        )



import re
from typing import Any, Container, List, Mapping, Optional, Union, overload

from maypy import Mapper, Maybe, Predicate
from maypy.predicates import (
    contains,
    equals,
    is_blank_str,
    is_empty,
    is_length,
    is_truthy,
    match_regex,
    neg,
    one_of,
)
from typing_extensions import Self, deprecated, override

from power_events.conditions.condition import And, Condition, Event, Or, V
from power_events.exceptions import NoPredicateError, ValueAbsentError


class Absent:
    """Sentinel class to represent value absence."""


ABSENT = Absent()


class ValuePath(str):
    """A path-like string for accessing nested mappings value."""

    SEPARATOR = "."
    separator: str
    keys: List[str]

    def __new__(cls, path: str, *, separator: Optional[str] = None) -> Self:
        """Create a new ValuePath object from a path string.

        Args:
            path: path string of a value inside a mapping object.
            separator: custom separator to use instead of the default.
        """
        separator = separator or cls.SEPARATOR
        is_blank = is_blank_str(path)

        if not is_blank:
            cls._validation(path, separator)

        instance = super().__new__(cls, path)
        instance.separator = separator
        instance.keys = [] if is_blank else instance.strip().split(separator)

        return instance

    def get_from(
        self,
        mapping: Mapping[Any, V],
        default: Union[V, None, Absent] = ABSENT,
        *,
        raise_if_absent: bool = False,
    ) -> Union[Any, None, Absent]:
        """Get the value describe by the path inside mapping object.

        The value returns can be the sentinel `ABSENT` to differentiate real `None` value of default.

        Args:
            mapping: Dictionary or mapping object to lookup.
            default: Default value if key not found. By default return the sentinel value `ABSENT`.
            raise_if_absent: Flag to raise `ValueAbsentError` if missing key. Default `False`

        Note:
            Support both string and integer keys.

        Raises:
            ValueAbsentError: if parameter `raise_if_absent` set, and key is missing.
        """
        value: Union[Mapping[Any, V], V, None, Absent] = dict(mapping)
        for key in self.keys:
            key_present = False

            if isinstance(value, Mapping):
                if key in value:
                    value = value[key]
                    key_present = True

                elif key.isdigit() and (num_key := int(key)) in value:
                    value = value[num_key]
                    key_present = True

            if not key_present:
                if raise_if_absent:
                    raise ValueAbsentError(self, key, mapping)
                return default
        return value

    @deprecated("""
    `get` is deprecated, use `get_from` instead.
    """)
    def get(
        self,
        mapping: Mapping[Any, V],
        default: Union[V, None, Absent] = ABSENT,
        *,
        raise_if_absent: bool = False,
    ) -> Union[Any, None, Absent]:
        """Get the value describe by the path inside mapping object.

        `get` is deprecated, use `get_from` instead.

        The value returns can be the sentinel `ABSENT` to differentiate real `None` value of default.

        Args:
            mapping: Dictionary or mapping object to lookup.
            default: Default value if key not found. By default return the sentinel value `ABSENT`.
            raise_if_absent: Flag to raise `ValueAbsentError` if missing key. Default `False`

        Note:
            Support both string and integer keys.

        Raises:
            ValueAbsentError: if parameter `raise_if_absent` set, and key is missing.
        """
        return self.get_from(mapping, default, raise_if_absent=raise_if_absent)

    @staticmethod
    def _validation(path: str, sep: str) -> None:
        """Validate string path."""
        if path[0] == sep:
            raise ValueError(f"Path value should not begin by key sep '{sep}'")

        if path[-1] == sep:
            raise ValueError(f"Path value should not end by key sep '{sep}'")

        if sep * 2 in path:
            raise ValueError(
                f"Path value should not contain consecutive separators, '{sep * 2}' is forbidden"
            )


class _MissingPredicate(Predicate[Any]):
    """A predicate that always returns False, representing a missing condition."""

    def __call__(self, val: Any) -> bool:
        raise NotImplementedError  # pragma: no cover

    def __repr__(self) -> str:
        return "Predicate<MISSING>"


MISSING = _MissingPredicate()


class Value(Condition):
    """Condition based on a value at a certain path in an event."""

    def __init__(self, value_path: str, mapper: Optional[Mapper[Any, Any]] = None) -> None:
        """Initialize the condition with the specified value path.

        Args:
            value_path: The path to the value in the event.
            mapper: mapper to transform value before checking predicates on it.
        """
        self.path: ValuePath = ValuePath(value_path)
        self._predicate: Predicate[Any] = MISSING
        self.mapper: Mapper[Any, Any] = mapper or (lambda val: val)

    @classmethod
    def root(cls) -> Self:
        """Initialize value targeting condition over the event itself.

        Examples:
            ```python
            Value.root().contains("a").check(my_dict)
            #  is equivalent to.
            "a" in my_dict
            ```
        """
        return cls("")

    @override
    def check(self, event: Event[V], *, raise_if_absent: bool = False) -> bool:
        """Check the given event respect the value condition.

        By default, the event fails the check if the value is absent, an error can be raised,
        by setting the flag `raise_if_absent`.

        Raises:
            NoPredicateError: when no predicate has been set.
            ValueAbsentError: if a key define by the path is missing in event.
        """
        if self._predicate is MISSING:
            raise NoPredicateError(self.path)

        if (val := self.path.get_from(event, raise_if_absent=raise_if_absent)) is ABSENT:
            return False

        return self._predicate(Maybe.of(val).map(self.mapper).or_else(val))

    def is_truthy(self) -> Self:
        """Add value is truthy check to the condition."""
        return self.__add(is_truthy)

    def equals(self, expected: Any) -> Self:
        """Add value equals the expected value check to the condition.

        Args:
            expected: The expected value.
        """
        return self.__add(equals(expected))

    @overload
    def match_regex(self, regex: re.Pattern[str]) -> Self: ...

    @overload
    def match_regex(self, regex: str, flags: Union[re.RegexFlag, int] = 0) -> Self: ...

    def match_regex(
        self, regex: Union[re.Pattern[str], str], flags: Union[re.RegexFlag, int] = 0
    ) -> Self:
        """Add value matches the given regex pattern check to the condition.

        Args:
            regex: regex to match (either a string or a Pattern)
            flags: regex flags; should bot be passed with a pattern.

        Raises:
            TypeError: when passing flags whereas a `Pattern` have been passed
        """
        return self.__add(match_regex(regex, flags))  # type: ignore[arg-type]

    def one_of(self, options: Container[Any]) -> Self:
        """Add value is one of the given options check to the condition.

        Args:
            options: The container of options.
        """
        return self.__add(one_of(options))

    def contains(self, *items: Any) -> Self:
        """Add value contains all the given items to the condition.

        Args:
            items: The items to check for.
        """
        return self.__add(contains(*items))

    def is_not_empty(self) -> Self:
        """Add value is not empty to the condition."""
        return self.__add(neg(is_empty))

    def is_length(self, length: int) -> Self:
        """Add value has the specified size to the condition.

        Args:
            length: The length expected.
        """
        return self.__add(is_length(length))

    def match(self, predicate: Predicate[Any]) -> Self:
        """Add value matches the given predicate to the condition.

        Args:
            predicate: The predicate to apply.
        """
        return self.__add(predicate)

    @override
    def __or__(self, other: Condition) -> Condition:
        if not isinstance(other, Condition):
            return NotImplemented
        return Or(self, other)

    @override
    def __and__(self, other: Condition) -> Condition:
        if not isinstance(other, Condition):
            return NotImplemented
        return And(self, other)

    @override
    def __invert__(self) -> Condition:
        invert = Value(self.path)
        invert._predicate = neg(self._predicate)
        return invert

    def __add(self, predicate: Predicate[Any]) -> Self:
        """Add new predicate to the current one.

        Args:
            predicate: The new predicate.
        """
        if self._predicate is MISSING:
            self._predicate = predicate
        else:
            self._predicate = combine(self._predicate, predicate)

        return self

    def __repr__(self) -> str:
        return f"Value(path={self.path}, predicate={self._predicate})"


def combine(*predicates: Predicate[Any]) -> Predicate[Any]:
    """Combine all provided predicates into a single predicate.

    Args:
        predicates: Predicates to combine.
    """

    def test(val: Any) -> bool:
        return all(predicate(val) for predicate in predicates)

    return test



from collections.abc import Mapping
from typing import Any, Dict, Union

import pytest

from power_events.conditions import Neg, Value
from power_events.exceptions import MultipleRoutesError, NoRouteFoundError
from power_events.resolver import EventResolver, EventRoute


class TestEventRoute:
    def test_match(self) -> None:
        def route(event: Any) -> int:
            return 1

        assert EventRoute(route, Value("a").equals(1)).match({"a": 1})

    def test_name(self) -> None:
        def route_test_name(event: Any) -> int:
            return 1

        assert EventRoute(route_test_name, Value("a").equals(1)).name == "route_test_name"
        assert EventRoute(lambda x: x, Value("a").equals(1)).name == "<lambda>"


class TestResolver:
    def test_resolve_should_raise_no_route_when_set_to_disallow(self) -> None:
        app = EventResolver(allow_no_route=False)

        with pytest.raises(NoRouteFoundError):
            app.resolve({"a": 1})

    def test_resolve_should_do_nothing_when_no_route(self) -> None:
        app = EventResolver()

        @app.equal("a.b", "TEST")
        def handle_test(_event: Mapping[str, Any]) -> str:
            return "lol"

        assert app.resolve({"a": 1}) == []

    def test_resolve_should_use_fallback_route_when_no_route_match(self) -> None:
        app = EventResolver()

        @app.equal("a.b", "TEST")
        def handle_test(_event: Mapping[str, Any]) -> str:
            return "lol"

        @app.fallback
        def fallback(_event: dict[str, str]) -> str:
            return "fallback"

        assert app.resolve({"a": 1}) == ["fallback"]

    def test_resolve_should_raise_multiple_route_when_set_to_disallow(self) -> None:
        app = EventResolver(allow_multiple_routes=False)

        @app.equal("a.b", "TEST")
        def handle_test(_event: Mapping[str, Any]) -> str:
            return "lol"

        @app.equal("a.b", "TEST")
        def handle_test2(_event: Dict[str, Any]) -> str:
            return "lol-2"

        with pytest.raises(MultipleRoutesError):
            app.resolve({"a": {"b": "TEST"}})

    def test_equal(self) -> None:
        app = EventResolver(allow_no_route=False)

        @app.equal("a.b", "TEST")
        def handle_test(_event: Dict[str, Any]) -> str:
            return "lol"

        res = app.resolve({"a": {"b": "TEST"}})

        assert res == ["lol"]

    def test_one_of(self) -> None:
        app = EventResolver()

        @app.one_of("a", ["BAR", "FOO"])
        def handle_foo(_event: Dict[str, Any]) -> str:
            return "BAR"

        res = app.resolve({"a": "FOO"})

        assert res == ["BAR"]

    def test_contain(self) -> None:
        app = EventResolver()

        @app.one_of("a", ["BAR", "FOO"])
        def handle_contain(_event: Dict[str, Any]) -> str:
            return "BAR"

        @app.contain("a", "b", "c")
        def handle_foo(_event: Dict[str, Any]) -> str:
            return "contain"

        res = app.resolve({"a": ["b", "c"]})

        assert res == ["contain"]

    def test_resolve_multiple_routes(self) -> None:
        event = {"a": {"b": {"c": "TEST"}, "d": 1}}
        app = EventResolver(allow_multiple_routes=True)

        @app.when(Value("a.b.c").one_of(["TEST"]))
        def handle_test(_event: Dict[str, Any]) -> str:
            return "test"

        @app.equal("a.d", 1)
        def handle_d(_event: Dict[str, Any]) -> str:
            return "d"

        res = app.resolve(event)

        assert res == ["test", "d"]

    def test_should_not_catch_exception_when_no_exception_handler(self) -> None:
        app = EventResolver()

        @app.one_of("a", ["BAR", "FOO"])
        def handle_contain(_event: Dict[str, Any]) -> str:
            return "BAR"

        @app.contain("a", "b", "c")
        def handle_foo(_event: Dict[str, Any]) -> str:
            raise ValueError("test")

        with pytest.raises(ValueError, match="test"):
            app.resolve({"a": ["b", "c"]})

    def test_should_not_catch_exception_when_no_exception_handler_for_this_error(self) -> None:
        app = EventResolver()

        @app.exception_handler(TypeError)
        def handle_value_error(exception: TypeError) -> str:
            return f"Something went wrong with value... {exception}"

        @app.contain("a", "b", "c")
        def handle_foo(_event: Dict[str, Any]) -> str:
            raise ValueError("test")

        with pytest.raises(ValueError, match="test"):
            app.resolve({"a": ["b", "c"]})

    def test_should_exception_handler_when_error_raised(self) -> None:
        app = EventResolver()

        @app.exception_handler(ValueError)
        def handle_value_error(exception: ValueError) -> str:
            return f"Something went wrong with value... {exception}"

        @app.one_of("a", ["BAR", "FOO"])
        def handle_contain(_event: Dict[str, Any]) -> str:
            return "BAR"

        @app.contain("a", "b", "c")
        def handle_foo(_event: Dict[str, Any]) -> str:
            raise ValueError("test")

        assert app.resolve({"a": ["b", "c"]}) == ["Something went wrong with value... test"]

    def test_should_have_registered_all_exceptions_has_handled_when_passing_exception_list(
        self,
    ) -> None:
        app = EventResolver()

        @app.exception_handler([ValueError, TypeError])
        def handle_value_error(exception: Union[ValueError, TypeError]) -> str:
            return f"Something went wrong with error of {type(exception).__name__}..."

        @app.one_of("a", ["BAR", "FOO"])
        def handle_contain(_event: Dict[str, Any]) -> str:
            raise TypeError()

        @app.contain("a", "b", "c")
        def handle_foo(_event: Dict[str, Any]) -> str:
            raise ValueError()

        assert app.resolve({"a": ["b", "c"]}) == [
            "Something went wrong with error of ValueError..."
        ]
        assert app.resolve({"a": "BAR"}) == ["Something went wrong with error of TypeError..."]

    def test_should_handle_exception_even_if_base_exception_registered(self) -> None:
        class CustomValueError(ValueError): ...

        app = EventResolver()

        @app.exception_handler(ValueError)
        def handle_value_error(exception: ValueError) -> str:
            return f"Something went wrong with error of {type(exception).__name__}..."

        @app.contain("a", "b", "c")
        def handle_foo(_event: Dict[str, Any]) -> str:
            raise CustomValueError()

        assert app.resolve({"a": ["b", "c"]}) == [
            "Something went wrong with error of CustomValueError..."
        ]

    def test_readme_simple(self) -> None:
        app = EventResolver()

        @app.equal("type", "order_created")
        def handler_order_created(event: dict[str, Any]) -> str:
            return f"Order created : {event['order_id']}"

        @app.one_of("type", ["order_update", "order_delete"])
        def handle_order_modification(event: dict[str, Any]) -> str:
            return f"Order modification <{event['type']}> : {event['order_id']}"

        assert app.resolve({"type": "order_created", "order_id": "12345", "user_id": "67890"}) == [
            "Order created : 12345"
        ]

        assert app.resolve({"type": "order_delete", "order_id": "12345", "user_id": "67890"}) == [
            "Order modification <order_delete> : 12345"
        ]

    def test_readme_complex(self) -> None:
        app = EventResolver()

        # Order created and digital purchase.
        @app.when(Value("type").equals("order_created") & Value("cart.is_digital").is_truthy())
        def handle_digital_purchase(event: dict[str, Any]) -> str:
            return f"The order created is a digital purchase: {event['order_id']}"

        # Order created and physical.
        @app.when(Value("type").equals("order_created") & Neg(Value("cart.is_digital").is_truthy()))
        def handle_physical_purchase(event: dict[str, Any]) -> str:
            return f"The order created is a physical purchase: {event['order_id']}"

        assert app.resolve(
            {
                "type": "order_created",
                "order_id": "12345",
                "user_id": "67890",
                "cart": {"is_digital": True, "items": ["$10 voucher"]},
            }
        ) == ["The order created is a digital purchase: 12345"]
        assert app.resolve(
            {
                "type": "order_created",
                "order_id": "12345",
                "user_id": "67890",
                "cart": {"is_digital": False, "items": ["keyboard"]},
            }
        ) == ["The order created is a physical purchase: 12345"]
